"""
delimited.container
~~~~~~~~~~~~~~~~~~~

This module defines NestedContainer objects. A NestedContainer object 
implements an interface through which nested data can be accessed and 
modified using Path objects.
"""

import abc
import copy
import collections

from delimited.path import TuplePath
from delimited.path import DelimitedStrPath


class NestedContainer(abc.ABC, dict):
    """ The abstract base class for NestedContainer objects. When subclassing 
    NestedContainer the path and container attributes must be overridden with a 
    Path object and container type respectively. The path object chosen defines 
    the external representation used by the NestedContainer class.
    """
    
    """ The Path object used to define paths to nested data
    """
    path = None

    """ The container type that this NestedContainer emulates
    """
    container = None

    def __init__(self, data=None):
        self(data)

    def __call__(self, data=None):
        """ Overwrite instance data with expanded data param. If data param is 
        None, set instance data to new instance of container attribute.
        """
        
        self.data = self.container() if data is None else self._expand(data)

    def __eq__(self, other):
        """ Compare equality with instance of self or instance of container 
        attribute.
        """

        if isinstance(other, self.__class__):
            return self.data == other.data
        return self.data == other

    def __ne__(self, other):
        """  Compare inequality with instance of self or instance of container 
        attribute.
        """
        return not self.__eq__(other)

    def __hash__(self):
        """ Return has of instance data.
        """
        
        return hash(str(self.data))

    def __bool__(self):
        """ Return instance data cast to boolean.
        """
        return bool(self.data)

    def __repr__(self):
        """ Return class name joined with instance data cast to str.
        """
        
        return f'{self.__class__.__name__}({self.data})'

    def __iter__(self):
        """ Yield key, value tuples for instance data.
        """
        
        for key, value in self.data.items():
            yield key, value

    def __contains__(self, path):
        """ Return boolean for presence of path in instance data. Accepts path 
        segment or value in collapsed path format.
        """
        
        return self.has(path)

    def __getitem__(self, path):
        """ Return a reference to the value at path. Raise exception if path
        cannot be resolved.
        """
        
        return self.ref(path)

    def __setitem__(self, path, value):
        """ Set data at path to value.
        """
        
        self.set(path, value)

    def __delitem__(self, path):
        """ Remove key and value at path. Raise exception if path cannot be 
        resolved.
        """
        
        self.unset(path)

    def __len__(self):
        """ Return length of first level of keys of instance data.
        """
        
        return len(self.data)

    def __copy__(self):
        """ Return new instance of self with its data set to a shallow copy of
        this instances data.
        """
        
        new = self.__class__()
        new.data = copy.copy(self.data)
        return new

    def __deepcopy__(self, *args):
        """ Return new instance of self with its data set to a deep copy of 
        this instances data.
        """
        
        new = self.__class__()
        new.data = copy.deepcopy(self.data)
        return new

    def items(self):
        """ Yield key, value tuples for instance data.
        """
        
        for key, value in self.data.items():
            yield (key, value)

    def keys(self):
        """ Yield keys for instance data.
        """
        
        for key in self.data.keys():
            yield key

    def values(self):
        """ Yield values for instance data.
        """
        
        for value in self.data.values():
            yield value

    def ref(self, path=None, create=False):
        """ Return a reference to nested data at path. If create is True and 
        missing key(s) are encountered while trying to resolve path, create 
        the missing key(s) using an instance of self.container as the value.
        """

        # setup haystack
        haystack = self.data

        # all attributes
        if path is None:
            return haystack

        # setup path
        if not isinstance(path, self.path):
            path = self.path(path)

        # for each needle
        for needle in path:
            
            try:
                haystack = haystack[needle]

            except KeyError:                
                if create:
                    haystack[needle] = self.container()
                    haystack = haystack[needle]
                    
                else:
                    # TODO: custom exception
                    raise

            except TypeError:
                # TODO: custom exception
                raise

        return haystack

    def get(self, path=None, *args):
        """ Return a copy of nested data at path. First value of args is 
        considered the default value, and if self.ref call raises a KeyError,
        the default value will be returned.
        """
        
        try:
            return copy.deepcopy(self.ref(path))
        except KeyError:
            if args:
                return args[0]
            # TODO: custom exception    
            raise
            
        except TypeError:
            # TODO: custom exception
            raise

    def has(self, path=None):
        """ Return True if path can be resolved in instance data else False.
        """
        
        try:
            return bool(self.ref(path))
        except KeyError:
            return False

    def spawn(self, path=None):
        """ Return a new instance of self with its data set to a reference of 
        this instances data.
        """
        
        spawn = self.__class__()
        spawn.data = self.ref(path)
        return spawn

    def clone(self, path=None):
        """ Return a new instance of self with its data set to a deep copy of 
        this instances data.
        """
        
        return self.__class__(self.get(path))

    @classmethod
    def _merge(cls, a, b):
        """ Recursively merge container a into a deep copy of container b
        """

        b = copy.deepcopy(b)
        for k in a.keys():
            if k in b and all(isinstance(c, cls.container) for c in [a[k], b[k]]):
                b[k] = cls._merge(a[k], b[k])
            else:
                b[k] = a[k]
        return b

    def merge(self, data, path=None):
        """ Merge data with a copy of instance data at path and return 
        merged data. Will accept instance of self or instance of 
        self.container.
        """

        if isinstance(data, self.__class__):
            data = data.data
        elif isinstance(data, self.container):
            data = self.__class__(data).data
        return self._merge(data, self.get(path))

    @classmethod
    def _expand(cls, data):
        """ Recursively expand collapsed nested data and return.
        """

        expanded = cls.container()
        for path, value in data.items():

            # handle path
            if not isinstance(path, cls.path):
                path = cls.path(path)

            # expand value
            if isinstance(value, cls.container):
                value = cls._expand(value)

            for i, key in enumerate(reversed(path), 1):

                # first key
                if i == 1:
                    expanded_key = cls.container({key: value})

                # after first key
                elif i > 1:
                    expanded_key = cls.container({key: expanded_key})

                # last key, merge with other keys
                if i == len(path):
                    expanded = cls._merge(expanded, expanded_key)

        return expanded

    @classmethod
    def _collapse(cls, data, _path=None, func=None):
        """ Recursively collapse expanded nested data and return.
        """

        collapsed = cls.container()
        path = _path.copy() if _path is not None else cls.path()

        for key, value in data.items():

            if callable(func) and func(key):
                if isinstance(value, cls.container) and len(value):
                    value = cls._collapse(value, func=func)
                value = {path.encode(): {key: value}}

            else:

                path.extend(key)
                if isinstance(value, cls.container) and len(value):
                    value = cls._collapse(value, _path=path, func=func)
                else:
                    value = {path.encode(): value}

            collapsed.update(value)

        return collapsed

    def collapse(self, path=None, func=None):
        """ Collapse instance data at path and return. Use func to determine 
        if a level of nested data should be collapsed or not.
        """
        
        return self._collapse(self.get(path), func=func)

    def update(self, data, path=None):
        """ Update instance data at path with data.
        """
        
        haystack = self.ref(path)

        if isinstance(data, self.__class__):
            data = data.data
        elif isinstance(data, self.container):
            data = self._expand(data)

        haystack.update(data)

    def set(self, path, value, create=True):
        """ Set value at path in instance data. If create is True, create 
        missing keys while trying to resolve path.
        """

        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path[:-1], create=create)
        needle = path[-1]
        haystack[needle] = value

    def push(self, path, value, create=True):
        """ Push value to list at path in instance data. If create is True and 
        key for final path segment is not set, create key and create value as 
        empty list and append value to list. If create is True and key for 
        final path segment is wrong type, create value as list with existing 
        value and append value to list.
        """

        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path[:-1], create=create)
        needle = path[-1]

        try:
            haystack[needle].append(value)

        except KeyError as e:
            if create:
                haystack[needle] = []
                haystack[needle].append(value)
            else:
                raise

        except AttributeError as e:
            if create:
                haystack[needle] = [haystack[needle]]
                haystack[needle].append(value)
            else:
                raise

        return True

    def pull(self, path, value, cleanup=False):
        """ Remove value from list at path in instance data. If cleanup is True
        and removal of value results in an empty list, remove list and key.
        """

        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path[:-1])
        needle = path[-1]

        haystack[needle].remove(value)

        if cleanup:
            if haystack[needle] == []:
                del haystack[needle]

        return True

    def unset(self, path, cleanup=False):
        """ Remove value and last key at path. If cleanup is True and key 
        removal results in empty container, recursively remove empty 
        containers in reverse order of path.
        """

        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path[:-1])
        needle = path[-1]

        del haystack[needle]

        if cleanup:
            for i, needle in enumerate(path, 1):
                if i < len(path):
                    cleanup_path = path[:(len(path) - i)]
                    if self.ref(cleanup_path) == self.container():
                        self.unset(cleanup_path)

        return True


class NestedDict(NestedContainer):
    """ This class implements tuple path notation in use with the dict 
    container type.
    """
    
    path = TuplePath
    container = dict


class NestedOrderedDict(NestedContainer):
    """ This class implements tuple path notation in use with the OrderedDict 
    container type.
    """
    
    path = TuplePath
    container = collections.OrderedDict


class DelimitedDict(NestedContainer):
    """ This class implements delimited string path notation in use with the 
    dict container type.
    """
    
    path = DelimitedStrPath
    container = dict


class DelimitedOrderedDict(NestedContainer):
    """ This class implements delimited string path notation in use with the 
    OrderedDict container type.
    """
    
    path = DelimitedStrPath
    container = collections.OrderedDict



import sys; sys.path.append("../")

import unittest
import copy

from delimited.path import TuplePath
from delimited.path import DelimitedStrPath
from delimited.container import NestedDict
from delimited.container import DelimitedDict


class TestNestedDict(unittest.TestCase):

    # __init__

    def test___init____no_params(self):
        a = NestedDict()
        self.assertEqual(a.data, {})

    def test___init____dict_param(self):
        a = NestedDict({("k",): "v"})
        self.assertEqual(a, {"k": "v"})

    def test___init____tuple_notation_dict_param(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertEqual(a, {"k1": {"k2": {"k3": "v"}}})

    # __call__

    def test___call____no_params(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        a()
        self.assertEqual(a, {})

    def test___call____dict_param(self):
        a = NestedDict()
        a({("k",): "v"})
        self.assertEqual(a, {"k": "v"})

    def test___call____tuple_notation_dict_param(self):
        a = NestedDict()
        a({("k1", "k2", "k3"): "v"})
        self.assertEqual(a, {"k1": {"k2": {"k3": "v"}}})

    # __bool__

    def test___bool____empty_dict__returns_False(self):
        self.assertFalse(bool(NestedDict()))

    def test___bool____non_empty_dict__returns_True(self):
        self.assertTrue(bool(NestedDict({("k1", "k2", "k3"): "v"})))

    # __repr__

    def test___repr__(self):
        a = NestedDict({("k",): "v"})
        self.assertEqual(repr(a), "NestedDict({'k': 'v'})")

    # __eq__

    def test___eq____returns_True(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertTrue(a1 == a2)

    def test___eq____same_class__returns_False(self):
        a1 = NestedDict({("foo", "bar"): "baz"})
        a2 = NestedDict({("bar", "baz"): "foo"})
        self.assertFalse(a1 == a2)

    def test___eq____different_class__returns_True(self):
        a1 = NestedDict({("foo", "bar"): "baz"})
        a2 = {"foo": {"bar": "baz"}}
        self.assertTrue(a1 == a2)

        a3 = NestedDict()
        a4 = {}
        self.assertTrue(a3 == a4)

    # __ne__

    def test___ne____returns_True(self):
        a1 = NestedDict({("foo", "bar"): "baz"})
        a2 = NestedDict({("bar", "baz"): "foo"})
        self.assertTrue(a1 != a2)

    def test___ne____returns_False(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertFalse(a1 != a2)

    # __hash__

    def test___hash__(self):
        a1 = hash(NestedDict({("k1", "k2", "k3"): "v"}))
        a2 = hash(NestedDict({("k1", "k2", "k3"): "v"}))
        self.assertTrue(a1 == a2)

    # __iter__

    def test___iter__(self):
        a = NestedDict({("a", "a", "a"): "v"})
        for k, v in a:
            self.assertEqual(k, "a")

    # __contains__

    def test___contains__(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertTrue(("k1",) in a)

    # __getitem__

    def test___getitem__tuple_param__returns_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertEqual(a[("k1",)], {"k2": {"k3": "v"}})

    def test___getitem__delimited_tuple_param__returns_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertEqual(a[("k1", "k2", "k3")], "v")

    # __setitem__

    def test___setitem__string_param__sets_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        a[("k1",)] = "foo"
        self.assertEqual(a.data, {"k1": "foo"})

    def test___setitem__delimited_string_param__sets_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        a[("k1", "k2", "k3")] = "foo"
        self.assertEqual(a.data, {"k1": {"k2": {"k3": "foo"}}})

    # __delitem__

    def test___delitem__string_param__deletes_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        del a[("k1",)]
        self.assertEqual(a.data, {})

    def test___delitem__delimited_string_param__deletes_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        del a[("k1", "k2")]
        self.assertEqual(a.data, {"k1": {}})

    # __len__

    def test___len__(self):
        a = NestedDict({("foo",): "v", ("bar",): "v", ("baz",): "v"})
        self.assertEqual(len(a), 3)

    # __copy__

    def test___copy__(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = copy.copy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.set(("k1", "k2"), "bar")
        self.assertEqual(a1, a2)

    # __deepcopy__

    def test___deepcopy__(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = copy.deepcopy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.set(["k1", "k2"], "bar")
        self.assertNotEqual(a1, a2)

    # __items__

    def test___items___for_item__yields_tuple(self):
        a = NestedDict({("k",): "v"})
        for item in a.items():
            self.assertIsInstance(item, tuple)

    # __keys__

    def test___keys___yields_keys(self):
        a = NestedDict({("k",): "v"})
        for k in a.keys():
            self.assertEqual(k, "k")

    # __values__

    def test___values___yields_values(self):
        a = NestedDict({("k",): "v"})
        for v in a.values():
            self.assertEqual(v, "v")

    # ref

    def test_ref__no_param__returns_all_attributes(self):
        a = NestedDict({("k",): "v"})
        self.assertEqual(a.ref(), {"k": "v"})

    def test_ref__string_param__returns_value(self):
        a = NestedDict({("k",): "v"})
        self.assertEqual(a.ref("k"), "v")

    def test_ref__string_param__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.ref("k")

    def test_ref__string_param__raises_ValueError(self):
        a = NestedDict({("k1", "k2"): "v"})
        with self.assertRaises(TypeError):
            a.ref(("k1", "k2", "k3"))

    def test_ref__delimited_string_param__returns_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertEqual(a.ref(("k1", "k2")), ({"k3": "v"}))

    def test_ref__True_create_param_key_missing__creates_missing_containers(self):
        a = NestedDict()
        a.ref("k", create=True)
        self.assertEqual(a, {"k": {}})

    # get

    def test_get__string_param__returns_value(self):
        a = NestedDict({"k": "v"})
        self.assertEqual(a.get("k"), "v")

    def test_get__string_param_missing_key__returns_default_value(self):
        a = NestedDict()
        self.assertEqual(a.get("k", "foo"), "foo")

    def test_get__string_param__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.get("k")

    def test_get__string_param__raises_ValueError(self):
        a = NestedDict({("k1", "k2"): "v"})
        with self.assertRaises(TypeError):
            a.get(["k1", "k2", "k3"])

    def test_get__delimited_string_param__returns_value(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertEqual(a.get(("k1", "k2")), {"k3": "v"})

    # has

    def test_has__no_params__returns_True(self):
        self.assertTrue(NestedDict({"k": "v"}).has())

    def test_has__no_params__returns_False(self):
        self.assertFalse(NestedDict().has())

    def test_has__tuple_param__returns_True(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertTrue(a.has(("k1",)))

    def test_has__tuple_param__returns_False(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertFalse(a.has(("foo",)))

    def test_has__delimited_tuple_param__returns_True(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertTrue(a.has(("k1", "k2", "k3")))

    def test_has__delimited_tuple_param__returns_False(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertFalse(a.has(("k1", "k2", "foo")))

    # spawn

    def test_spawn(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = a1.spawn()
        self.assertEqual(a1, a2)
        a1["k1"] = "foo"
        self.assertEqual(a1, a2)

    # clone

    def test_clone(self):
        a1 = NestedDict({("k1", "k2", "k3"): "v"})
        a2 = a1.clone()
        self.assertEqual(a1, a2)
        a1["k1"] = "foo"
        self.assertNotEqual(a1, a2)

    # set

    def test_set__string_key_param(self):
        a = NestedDict()
        a.set("k", "v")
        self.assertEqual(a, {"k": "v"})

    def test_set__delimited_string_key_param(self):
        a = NestedDict()
        a.set(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": "v"}}})

    def test_set__raises_TypeError(self):
        a = NestedDict({("k1", "k2"): "v"})
        with self.assertRaises(TypeError):
            a.set(("k1", "k2", "k3"), "v", create=False)

    # push

    def test_push__string_key_param(self):
        a = NestedDict({"k": []})
        a.push(("k",), "v")
        self.assertEqual(a, {"k": ["v"]})

    def test_push__string_key_param__convert_existing_value(self):
        a = NestedDict({"k": "v"})
        a.push(("k",), "v")
        self.assertEqual(a, {"k": ["v", "v"]})

    def test_push__delimited_string_key_param(self):
        a = NestedDict({("k1", "k2", "k3"): []})
        a.push(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": ["v"]}}})

    def test_push__True_create_param__creates_list(self):
        a = NestedDict()
        a.push(("k",), "v")
        self.assertEqual(a, {"k": ["v"]})

    def test_push__create_False__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.push(("k",), "v", create=False)

    def test_psuh__create_False__raises_TypeError(self):
        a = NestedDict({"k": "v"})
        with self.assertRaises(AttributeError):
            a.push(("k",), "v", create=False)

    # pull

    def test_pull__string_key_param(self):
        a = NestedDict({("k",): ["v"]})
        a.pull(("k",), "v")
        self.assertEqual(a, {"k": []})

    def test_pull__cleanup_True__removes_empty_containers(self):
        a = NestedDict({("k",): ["v"]})
        a.pull(("k",), "v", cleanup=True)
        self.assertEqual(a, {})

    def test_pull__delimited_string_key_param(self):
        a = NestedDict({("k1", "k2", "k3"): ["v"]})
        a.pull(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": []}}})

    def test_pull__string_key_param__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.pull(("k",), "v")

    def test_pull__string_key_param__raises_ValueError(self):
        a = NestedDict({"k": []})
        with self.assertRaises(ValueError):
            a.pull(("k",), "v")

    def test_pull__string_key_param__raises_TypeError(self):
        a = NestedDict({("k",): "v"})
        with self.assertRaises(AttributeError):
            a.pull(("k",), "v")

    # unset

    def test_unset__string_key_param(self):
        a = NestedDict({("k",): "v"})
        a.unset(("k",))
        self.assertEqual(a, {})

    def test_unset__cleanup_True__removes_empty_containers(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        a.unset(("k1", "k2", "k3"), cleanup=True)
        self.assertEqual(a, {})

    def test_unset__string_key_param__raises_KeyError(self):
        a = NestedDict({("k",): "v"})
        with self.assertRaises(KeyError):
            a.unset(("foo",))

    # _merge

    def test__merge__dict_params(self):
        a1 = {("k1",): "v"}
        a2 = {("k2",): "v"}
        a = NestedDict._merge(a1, a2)
        self.assertEqual(type(a), dict)
        self.assertEqual(a, {
            ("k1",): "v",
            ("k2",): "v"
        })

    def test__merge__nested_dict_params(self):
        a1 = {("foo",): "v", ("bar",): {("baz",): "v"}}
        a2 = {("bar",): {("qux",): "v"}}
        a = NestedDict._merge(a1, a2)
        self.assertEqual(type(a), dict)
        self.assertEqual(a, {
            ("foo",): "v",
            ("bar",): {
                ("baz",): "v",
                ("qux",): "v"
            }
        })

    def test__merge__NestdDict_params(self):
        a1 = NestedDict({("k1",): "v"})
        a2 = NestedDict({("k2",): "v"})
        a = NestedDict._merge(a1, a2)
        self.assertEqual(type(a), NestedDict)
        self.assertEqual(a, NestedDict({
            ("k1",): "v",
            ("k2",): "v"
        }))

    # merge

    def test_merge__dict_param(self):
        a = NestedDict({("k1",): "v"})

        b = a.merge({("k2",): "v"})
        self.assertEqual(b, {
            "k1": "v",
            "k2": "v"
        })

    def test_merge__NestedDict_param(self):
        a = NestedDict({("k1",): "v"})
        b = a.merge(NestedDict({("k2",): "v"}))
        self.assertEqual(b, {
            "k1": "v",
            "k2": "v"
        })

    # update

    def test_update__dict_param(self):
        a = NestedDict({("k1",): "v"})
        a.update({("k2",): "v"})
        self.assertEqual(a, {
            "k1": "v",
            "k2": "v"
        })

    def test_update__NestedDict_param(self):
        a = NestedDict({("k1",): "v"})
        a.update(NestedDict({("k2",): "v"}))
        self.assertEqual(a, {
            "k1": "v",
            "k2": "v"
        })

    # _collapse

    def test__collapse(self):
        data = {("k1",): {("k2",): {("k3",): "v"}}}
        b = NestedDict()
        b.data = {("k1", "k2", "k3"): "v"}
        self.assertEqual(NestedDict._collapse(data), b)

    def test__collapse__empty_dict(self):
        data = {("k1",): {("k2",): {("k3",): {}}}}
        b = NestedDict()
        b.data = {("k1", "k2", "k3"): {}}
        self.assertEqual(NestedDict._collapse(data), b)

    # collapse

    def test_collapse(self):
        a = NestedDict({("k1",): {("k2",): {("k3",): "v"}}})
        b = a.collapse()
        self.assertEqual(b, {("k1", "k2", "k3"): "v"})

    def test_collapse__function_param(self):
        a = NestedDict({("k1", "k2", "$foo", "k3", "k4"): "v"})

        def detect_mongo_operator(path):
            return path[0][0] == "$"

        b = a.collapse(func=detect_mongo_operator)
        self.assertEqual(b, {("k1", "k2"): {"$foo": {("k3", "k4"): "v"}}})

    # _expand

    def test__expand(self):
        a = {("k1", "k2", "k3"): "v"}
        b = NestedDict._expand(a)
        self.assertEqual(b, {"k1": {"k2": {"k3": "v"}}})

    def test__expand__overlapping_paths(self):
        a = {
            ("k1", "k2", "k3"): "v",
            ("k1", "k2", "k4"): "v"
        }
        b = NestedDict._expand(a)
        self.assertEqual(b, {"k1": {"k2": {"k3": "v", "k4": "v"}}})

    def test__expand__nested_class(self):
        a = {("k1",): NestedDict({("k2", "k3"): "v"})}
        b = NestedDict._expand(a)
        self.assertEqual(b, {"k1": {"k2": {"k3": "v"}}})


if __name__ == "__main__":
    unittest.main()



__version__ = "0.0.2"
__title__ = "delimited"
__summary__ = "Defines types that allow for accessing and modifying nested data"
__url__ = "https://github.com/chrisantonellis/delimited"
__author__ = "Christopher Antonellis"
__email__ = "christopher.antonellis@gmail.com"
__license__ = "MIT License"

from delimited.container import NestedDict
from delimited.container import NestedOrderedDict
from delimited.container import DelimitedDict
from delimited.container import DelimitedOrderedDict
